//  Copyright (c) 2012 Dennco Project
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

//
//  Created by tkawata on 1/6/2012.
//

#include "TKLock.h"
#include "TKContainer.h"
#include "DNUtils.h"
#include "DNDirectory.h"
#include "DNFileList.h"
#include "DNXML.h"
#include "DNXMLElement.h"
#include "TKLog.h"
#include "DNUtils.h"
#include "TKCell.h"
#include "TKDebug.h"

#include "DNContainerBuilder.h"

DNContainerBuilder::DNContainerBuilder(TKContainer *container) : mContainer(container) 
{
    
}

DNContainerBuilder::~DNContainerBuilder()
{
    
}

TKContainer* DNContainerBuilder::buildContainerFromXHTML(const char* docRoot)
{

    mLock.lock();
    
    mContainer->setContentPath(std::string(docRoot));

    DNDirectory directory(docRoot);
    const DNFileList *flist = directory.getFileList("html|htm|xml");
    
    //parse XHTML files
    while(flist)
    {
        parseXHTML(docRoot, flist->filePath.c_str());

        flist = flist->next;
    }
    
    //build cells
    while(!mPendingCellDefinitions.empty())
    {
        PendingCellDefinition *cellInfo = mPendingCellDefinitions.front();
        TKCellCode *cellCode = mContainer->getCellCode(cellInfo->cellCodeName);
        TKCell *newCell = NULL;
        if (cellCode)
        {
            newCell = mContainer->createCell(cellInfo->location, cellInfo->cellName, cellCode, cellInfo->startupScript);
        }
        if (newCell == NULL)
        {
            TKLog::printf("Failed to construct cell:%s.",cellInfo->cellName.c_str());
        }
        mPendingCellDefinitions.pop();
    }
    
    //build connections
    while(!mPendingConnections.empty())
    {
        PendingCellConnection *connInfo = mPendingConnections.front();
        TKCell *orgCell = mContainer->getCell(connInfo->originCellName);
        TKCell *tgtCell = mContainer->getCell(connInfo->targetCellName);
        
        if (orgCell && tgtCell)
        {
            orgCell->connectTo(connInfo->connectionName, tgtCell);
        }
        else
        {
            TKLog::printf("Failed to make connection (%s) cell:%s -> cell:%s", connInfo->connectionName.c_str(),connInfo->originCellName.c_str(), connInfo->targetCellName.c_str());            
        }
        mPendingConnections.pop();
    }
    
    mLock.unlock();
    
    return mContainer;
}

void DNContainerBuilder::parseXHTML(const char *docRoot, const char *path)
{
    
    DNXML *xml = DNXML::createXMLFromFile(docRoot, path);
    DNXMLElement *element = xml->getRoot();
    
    DNXMLElement *classElement = NULL;
    DNXMLElement *classScriptElement = NULL;
    DNXMLElement *classAPIElement = NULL;
    std::string  className;

    DNXMLElement *cellElement = NULL;
    DNXMLElement *cellScriptElement = NULL;
    std::string  cellName;
    std::string  cellCodeClassName;
    
    while(element)
    {
        std::string eclass = element->getAttributeValue("class");
        if (eclass == "dennco-script")
        {
            DNXMLElement *te = element->outer;
            while(te)
            {
                std::string teclass = te->getAttributeValue("class");
                if (teclass == "dennco-classdef")
                {
                    if (classElement && classElement == te)
                    {
                        classScriptElement = element;
                    }
                    else
                    {                    
                        TKLog::printf("A syntax error is found in %s. dennco-script isn't properly defined.",path);
                    }
                    
                    break;
                }
                else if (teclass == "dennco-cell")
                {
                    if (cellElement && cellElement == te)
                    {
                        std::string s = element->getAttributeValue("href");
                        if (s.length() > 0)
                        {
                            cellCodeClassName = s;
                        }
                        else
                        {
                            cellScriptElement = element;
                        }
                    }
                    else
                    {
                        TKLog::printf("A syntax error is found in %s. dennco-script isn't properly defined.",path);
                    }
                    
                    break;
                }
                te = te->outer;
            }
            if (!te)
            {
                TKLog::printf("A syntax error is found in %s. dennco-script doesn't belong to any of dennco-classdef or dennco-cell.", path);
            }
            
        }
        else if (eclass == "dennco-classdef")
        {            
            className = element->getAttributeValue("name");
            if (className.length() > 0)
            {
                if (!classElement)
                {
                    classElement = element;
                }
                else
                {
                    TKLog::printf("A syntax error is found in %s. Nested dennco-classdef?", path);
                }
            }
            else
            {
                TKLog::printf("A syntax error is found in %s. dennco-classdef doesn't have name attribute", path);                
            }
        }
        else if (eclass == "dennco-cell")
        {
            cellName = element->getAttributeValue("name");
            trimString(cellName);
            if (cellName.length() > 0)
            {
                if (!cellElement)
                {
                    cellElement = element;
                }
                else
                {
                    TKLog::printf("A syntax error is found in %s. Nested dennco-cell?", path);
                }
            }
            else
            {
                TKLog::printf("A syntax error is found in %s. dennco-cell doesn't have name attribute", path);
            }
        }
        else if (eclass == "dennco-connect")
        {
            if (cellElement)
            {
                //connection
                std::string target = element->getAttributeValue("href");
                trimString(target);
                if (target.length() > 0)
                {
                    PendingCellConnection *newConnection = new PendingCellConnection();
                    newConnection->connectionName = getFQNString(path, cellName.c_str());
                    newConnection->originCellName = getFQNString(path, cellName.c_str());
                    newConnection->targetCellName = getFQNString(path, target.c_str());
                    mPendingConnections.push(newConnection);
                }
                else
                {
                    TKLog::printf("A syntax error is found in %s. dennco-connect doesn't have link target (href) attribute.", path);                    
                }
            }
            else
            {
                TKLog::printf("A syntax error is found in %s. dennco-connect doesn't belong to any of dennco-cell.", path);
            }
        }
        else if (eclass == "dennco-api")
        {
            DNXMLElement *te = element->outer;
            while(te)
            {
                std::string teclass = te->getAttributeValue("class");
                if (teclass == "dennco-classdef")
                {
                    if (classElement && classElement == te)
                    {
                        classAPIElement = element;
                    }
                    else
                    {
                        TKLog::printf("A syntax error is found in %s. dennco-api isn't properly defined.",path);
                    }
                    
                    break;
                }
                te = te->outer;
            }
            if (!te)
            {
                TKLog::printf("A syntax error is found in %s. dennco-api doesn't belong to any of dennco-cell.", path);
            }
        }
        
        //go to next element
        //
        if (element->inner)
        {
            // go inside
            element = element->inner;
        }
        else
        {
            // go next 
            do {
                if (!element->next)
                {
                    element = element->outer;
                }

                // the processed element was cell code class definition.
                // define it now.
                if (classElement && (element == NULL || element->depth <= classElement->depth))
                {
                    std::string _cellCodeName = getFQNString(path,className.c_str());
                    std::string _cellCodeAPIType = classAPIElement->text;
                    std::string _cellCodeScript = classScriptElement->text;
                    
                    DEBUG_TRACE("\n===== Define cell code class ===== \nClass:%s\nAPI:%s\n%s\n", _cellCodeName.c_str(), _cellCodeAPIType.c_str(), _cellCodeScript.c_str());
                    
                    mContainer->createCellCode(_cellCodeName, _cellCodeAPIType, _cellCodeScript);
                    
                    
                    classElement = NULL;
                    classScriptElement = NULL;
                    classAPIElement = NULL;
                    className = "";
                }
            
                // the processed element was cell definition.
                // define it now.
                if (cellElement && (element == NULL || element->depth <= cellElement->depth))
                {
                    //scriptless celldef                    
                    std::string _cellName = cellName;
                    std::string _cellNameFQN = getFQNString(path, cellName.c_str());
                    std::string _cellCodeName = getFQNString(path, cellCodeClassName.c_str());
                    std::string _cellScript = cellScriptElement->text;
                    
                    DEBUG_TRACE("\n===== Define cell ================ \nName:%s\nClass:%s\nStartup script:\n%s\n", _cellNameFQN.c_str(), _cellCodeName.c_str(),_cellScript.c_str());
                    
                    
                    PendingCellDefinition *newCellDefine = new PendingCellDefinition();
                    newCellDefine->location = path;
                    newCellDefine->cellName = _cellName;
                    newCellDefine->cellCodeName = _cellCodeName;
                    newCellDefine->startupScript = _cellScript;
                    
                    mPendingCellDefinitions.push(newCellDefine);
                    
                    cellElement = NULL;
                    cellScriptElement = NULL;
                    cellName = "";
                    cellCodeClassName = "";
                }
                
            } while(element && !element->next && element != xml->getRoot());
            
            if (!element)
                break;
            element = element->next;
        }
    }
    
}